"""URL filter."""
from .. import filters
import re

# Bare link/email detection
RE_MAIL = re.compile(
    r'''(?xi)
    (?P<mail>
        (?<![-/\+@a-z\d_])(?:[-+a-z\d_]([-a-z\d_+]|\.(?!\.))*)  # Local part
        (?<!\.)@(?:[-a-z\d_]+\.)                                # @domain part start
        (?:(?:[-a-z\d_]|(?<!\.)\.(?!\.))*)[a-z]\b               # @domain.end (allow multiple dot names)
        (?![-@])                                                # Don't allow last char to be followed by these
    )
    '''
)

RE_LINK = re.compile(
    r'''(?xi)
    (?P<link>
        (?:(?<=\b)|(?<=_))(?:
            (?:ht|f)tps?://(?:(?:[^_\W][-\w]*(?:\.[-\w.]+)+)|localhost)|  # (http|ftp)://
            (?P<www>w{3}\.)[^_\W][-\w]*(?:\.[-\w.]+)+                     # www.
        )
        /?[-\w.?,!'(){}\[\]/+&@%$#=:"|~;]*                                # url path, fragments, and query stuff
        (?:[^_\W]|[-/#@$+=])                                              # allowed end chars
    )
    '''
)


class URLFilter(filters.Filter):
    """URL filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "emails": True,
            "urls": True
        }

    def setup(self):
        """Setup."""

        self.emails = self.config['emails']
        self.urls = self.config['urls']

    def _filter(self, text):
        """Filter out the URL and email addresses."""

        if self.urls:
            text = RE_LINK.sub('', text)
        if self.emails:
            text = RE_MAIL.sub('', text)
        return text

    def filter(self, source_file, encoding):  # noqa A001
        """Open and filter the file from disk."""

        with open(source_file, 'r', encoding=encoding, errors='strict') as f:
            text = f.read()

        return [filters.SourceText(self._filter(text), source_file, encoding, 'url-free')]

    def sfilter(self, source):
        """Execute filter."""

        return [filters.SourceText(self._filter(source.text), source.context, source.encoding, 'url-free')]


def get_plugin():
    """Return the filter."""

    return URLFilter



"""Stylesheets filter."""
import re
import textwrap
from .. import filters

TARGETS = r'''(?x)
(?P<comments>
    (?P<block>/\*[^*]*\*+(?:[^/*][^*]*\*+)*/) # multi-line comments
    {}
) |
(?P<strings>
    "(?:\\.|[^"\\\n])*" |                     # double quotes
    '(?:\\.|[^'\\\n])*'                       # single quotes
) |
(?P<code>
    .[^/"' \t\n]*?                            # everything else
)
'''

INLINE_COMMENTS = r'| (?P<start>^)?(?P<leading_space>[ \t]*)?(?P<line>//(?:[^\n])*)'

RE_CSS = re.compile(TARGETS.format(''), re.DOTALL | re.MULTILINE)
RE_SCSS = re.compile(TARGETS.format(INLINE_COMMENTS), re.DOTALL | re.MULTILINE)

CSS = 0
SASS = 1
SCSS = 2

STYLESHEET_TYPE = {
    "css": CSS,
    "sass": SASS,
    "scss": SCSS
}


class StylesheetsFilter(filters.Filter):
    """Stylesheets filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "block_comments": True,
            "line_comments": True,
            "group_comments": False,
            "stylesheets": 'css'
        }

    def validate_options(self, k, v):
        """Validate options."""

        super().validate_options(k, v)
        if k == 'stylesheets' and v not in STYLESHEET_TYPE:
            raise ValueError(f"{self.__class__.__name}: '{v}' is not a valid value for '{k}'")

    def setup(self):
        """Setup."""

        self.blocks = self.config['block_comments']
        self.lines = self.config['line_comments']
        self.group_comments = self.config['group_comments']
        # If the style isn't found, just go with CSS, then use the appropriate prefix.
        self.stylesheets = STYLESHEET_TYPE.get(self.config['stylesheets'].lower(), CSS)
        self.prefix = next(k for k, v in STYLESHEET_TYPE.items() if v == SASS)
        self.pattern = RE_CSS if self.stylesheets == CSS else RE_SCSS

    def evaluate_block(self, groups):
        """Evaluate block comments."""

        if self.blocks:
            self.block_comments.append([groups['block'][2:-2], self.line_num, self.current_encoding])

    def evaluate_inline_tail(self, groups):
        """Evaluate inline comments at the tail of source code."""

        if self.lines:
            self.line_comments.append([groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding])

    def evaluate_inline(self, groups):
        """Evaluate inline comments on their own lines."""

        # Consecutive lines with only comments with same leading whitespace
        # will be captured as a single block.
        if self.lines:
            if (
                self.group_comments and
                self.line_num == self.prev_line + 1 and
                groups['leading_space'] == self.leading
            ):
                self.line_comments[-1][0] += '\n' + groups['line'][2:].replace('\\\n', '')
            else:
                self.line_comments.append(
                    [groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding]
                )
            self.leading = groups['leading_space']
            self.prev_line = self.line_num

    def evaluate(self, m):
        """Search for comments."""

        g = m.groupdict()
        if g["strings"]:
            self.line_num += g['strings'].count('\n')
        elif g["code"]:
            self.line_num += g["code"].count('\n')
        else:
            if g['block']:
                self.evaluate_block(g)
            elif self.stylesheets != CSS:
                if g['start'] is None:
                    self.evaluate_inline_tail(g)
                else:
                    self.evaluate_inline(g)
            self.line_num += g['comments'].count('\n')

    def extend_src_text(self, content, context, text_list, category):
        """Extend the source text list with the gathered text data."""

        prefix = self.prefix + '-' if self.prefix else ''

        for comment, line, encoding in text_list:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    prefix + category
                )
            )

    def extend_src(self, content, context):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, 'block-comment')
        self.extend_src_text(content, context, self.line_comments, 'line-comment')

    def find_content(self, text):
        """Find content."""

        for m in self.pattern.finditer(self.norm_nl(text)):
            self.evaluate(m)

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        content = []
        self.current_encoding = encoding
        self.line_num = 1
        self.prev_line = -1
        self.leading = ''
        self.block_comments = []
        self.line_comments = []

        self.find_content(text)
        self.extend_src(content, context)

        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse stylesheet file."""

        with open(source_file, 'r', encoding=encoding, errors='strict') as f:
            text = f.read()

        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Get filter."""

    return StylesheetsFilter



"""Main."""
import sys
import argparse
from pyspelling import spellcheck, __version__


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='pyspelling', description='Spell checking tool.')
    # Flag arguments
    parser.add_argument('--version', action='version', version=('%(prog)s ' + __version__))
    parser.add_argument('--debug', action='store_true', default=False, help=argparse.SUPPRESS)
    parser.add_argument('--verbose', '-v', action='count', default=0, help="Verbosity level.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--name', '-n', action='append', help="Specific spelling task by name to run.")
    group.add_argument('--group', '-g', action='append', help="Specific spelling task group to run.")
    parser.add_argument('--binary', '-b', action='store', default='', help="Provide path to spell checker's binary.")
    parser.add_argument(
        '--jobs', '-j',
        action='store',
        type=int,
        default=None,
        help=(
            "Specify the number of spell checker processes to run in parallel. "
            "Using 0 will utilize the maximum number of cores."
        )
    )
    parser.add_argument('--config', '-c', action='store', default='', help="Spelling config.")
    parser.add_argument(
        '--source', '-S',
        action='append',
        help="Specify override file pattern. Only applicable when specifying exactly one --name."
    )
    parser.add_argument(
        '--spellchecker', '-s', action='store', default='', help="Choose between aspell and hunspell."
    )
    parser.add_argument(
        '--skip-dict-compile',
        '-x',
        action='store_true',
        help="Skip dictionary compilation if the compiled file already exists."
    )
    args = parser.parse_args()

    return run(
        args.config,
        names=args.name,
        groups=args.group,
        binary=args.binary,
        spellchecker=args.spellchecker,
        sources=args.source,
        verbose=args.verbose,
        debug=args.debug,
        jobs=args.jobs,
        skip_dict_compile=args.skip_dict_compile
    )


def run(config, **kwargs):
    """Run."""

    names = kwargs.get('names', [])
    groups = kwargs.get('groups', [])
    binary = kwargs.get('binary', '')
    spellchecker = kwargs.get('spellchecker', '')
    verbose = kwargs.get('verbose', 0)
    sources = kwargs.get('sources', [])
    debug = kwargs.get('debug', False)
    jobs = kwargs.get('jobs', None)
    if jobs is not None and jobs < 0:
        jobs = 1
    skip_dict_compile = kwargs.get('skip_dict_compile', False)

    fail = False
    count = 0
    for results in spellcheck(
        config,
        names=names,
        groups=groups,
        binary=binary,
        checker=spellchecker,
        sources=sources,
        verbose=verbose,
        debug=debug,
        jobs=jobs,
        skip_dict_compile=skip_dict_compile
    ):
        count += 1
        if results.error:
            fail = True
            print(f'ERROR: {results.context} -- {results.error}')
        elif results.words:
            fail = True
            print(f'Misspelled words:\n<{results.category}> {results.context}')
            print('-' * 80)
            for word in results.words:
                print(word)
            print('-' * 80)
            print('')

    if fail:
        print('!!!Spelling check failed!!!')
    else:
        print('Spelling check passed :)')

    return fail


if __name__ == "__main__":
    sys.exit(main())



"""JavaScript filter."""
import re
import textwrap
from .. import filters

RE_JSDOC = re.compile(r"(?s)^/\*\*$(.*?)[ \t]*\*/", re.MULTILINE)

RE_BLOCK_COMMENT = re.compile(r'/\*[^*]*\*+(?:[^/*][^*]*\*+)*/', re.DOTALL | re.MULTILINE)
RE_COMMENT = re.compile(r'(?P<start>^)?(?P<leading_space>[ \t]*)?(?P<line>//(?:[^\n])*)', re.DOTALL | re.MULTILINE)
RE_STRING = re.compile(r'''"(?:\\.|[^"\\\n])*"|(?:\\.|[^'\\\n])*''', re.DOTALL | re.MULTILINE)
RE_TEMPLATE_START = re.compile(r'`((?:\\.|\$(?!\{)|[^`\\$])*)(`|\$\{)', re.DOTALL | re.MULTILINE)
RE_TEMPLATE_MIDDLE_END = re.compile(r'((?:\\.|\$(?!\{)|[^`\\$])*)(`|\$\{)', re.DOTALL | re.MULTILINE)

BACK_SLASH_TRANSLATION = {
    "\\b": '\b',
    "\\f": '\f',
    "\\r": '\r',
    "\\t": '\t',
    "\\n": '\n',
    "\\v": '\v',
    "\\\\": '\\',
    '\\"': '"',
    "\\'": "'",
    "\\\n": '',
    "\\`": "`",
    "\\0": "\000"
}


RE_ESC = re.compile(
    r'''(?x)
    (?P<oct>\\(?:[1-7][0-7]{0,2}|[0-7]{2,3}))|
    (?P<special>\\['"bfrt0nv\\\n])|
    (?P<char>\\u(?:\{[\da-fA-F]+\}|[\da-fA-F]{4})|\\x[\da-fA-F]{2}) |
    (?P<other>\\.)
    '''
)

RE_TEMP_ESC = re.compile(
    r'''(?x)
    (?P<special>\\['"`bfrt0nv\\\n])|
    (?P<char>\\u(?:\{[\da-fA-F]+\}|[\da-fA-F]{4})|\\x[\da-fA-F]{2}) |
    (?P<other>\\.)
    '''
)

RE_SURROGATES = re.compile(r'([\ud800-\udbff])([\udc00-\udfff])')


class JavaScriptFilter(filters.Filter):
    """JavaScript filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "block_comments": True,
            "line_comments": True,
            "group_comments": False,
            "decode_escapes": True,
            "strings": False,
            "jsdocs": False
        }

    def setup(self):
        """Setup."""

        self.blocks = self.config['block_comments']
        self.lines = self.config['line_comments']
        self.group_comments = self.config['group_comments']
        self.jsdocs = self.config['jsdocs']
        self.decode_escapes = self.config['decode_escapes']
        self.strings = self.config['strings']
        self.prefix = 'js'

    def replace_escapes(self, m):
        """Replace escapes."""

        groups = m.groupdict()
        esc = m.group(0)
        if groups.get('special'):
            value = BACK_SLASH_TRANSLATION[esc]
        elif groups.get('char'):
            try:
                if esc.endswith('}'):
                    value = chr(int(esc[3:-1], 16))
                else:
                    value = chr(int(esc[2:], 16))
            except Exception:
                value = esc
        elif groups.get('oct'):
            # JavaScript only supports hex range.
            # So \400 would be \40 + '0'
            value = int(esc[1:], 8)
            overflow = ''
            if value > 255:
                value = esc[1:-1]
                overflow + esc[-1]
            value = chr(value) + overflow
        elif groups.get('other'):
            value = esc[1:]
        return value.replace('\x00', '\n')

    def replace_surrogates(self, m):
        """Replace surrogates."""

        high, low = ord(m.group(1)), ord(m.group(2))
        return chr((high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000)

    def evaluate_strings(self, string, temp=False):
        """Evaluate strings."""

        value = ''
        if self.strings:
            if self.decode_escapes:
                value = RE_SURROGATES.sub(
                    self.replace_surrogates,
                    (RE_TEMP_ESC if temp else RE_ESC).sub(self.replace_escapes, string)
                )
            else:
                value = string
            if not temp:
                self.quoted_strings.append([value, self.line_num, 'utf-8'])
        return value

    def evaluate_inline_tail(self, groups):
        """Evaluate inline comments at the tail of source code."""

        if self.lines:
            self.line_comments.append([groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding])

    def evaluate_inline(self, groups):
        """Evaluate inline comments on their own lines."""

        # Consecutive lines with only comments with same leading whitespace
        # will be captured as a single block.
        if self.lines:
            if (
                self.group_comments and
                self.line_num == self.prev_line + 1 and
                groups['leading_space'] == self.leading
            ):
                self.line_comments[-1][0] += '\n' + groups['line'][2:].replace('\\\n', '')
            else:
                self.line_comments.append(
                    [groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding]
                )
            self.leading = groups['leading_space']
            self.prev_line = self.line_num

    def evaluate_block(self, comments):
        """Evaluate block comments."""

        if self.jsdocs:
            m1 = RE_JSDOC.match(comments)
            if m1:
                lines = []
                for line in m1.group(1).splitlines(True):
                    l = line.lstrip()
                    lines.append(l[1:] if l.startswith('*') else l)
                self.jsdoc_comments.append([''.join(lines), self.line_num, self.current_encoding])
            elif self.blocks:
                self.block_comments.append([comments[2:-2], self.line_num, self.current_encoding])
        elif self.blocks:
            self.block_comments.append([comments[2:-2], self.line_num, self.current_encoding])

    def extend_src_text(self, content, context, text_list, category):
        """Extend the source text list with the gathered text data."""

        prefix = self.prefix + '-' if self.prefix else ''

        for comment, line, encoding in text_list:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    prefix + category
                )
            )

    def extend_src(self, content, context):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, 'block-comment')
        self.extend_src_text(content, context, self.line_comments, 'line-comment')
        self.extend_src_text(content, context, self.jsdoc_comments, 'docs')
        self.extend_src_text(content, context, self.quoted_strings, 'strings')

    def find_content(self, text, index=0, backtick=False):
        """Find content."""

        curly_count = 0
        last = '\n'
        self.lines_num = 1
        length = len(text)
        while index < length:
            start_index = index
            c = text[index]
            if c == '{' and backtick:
                curly_count += 1
            elif c == '}' and backtick:
                if curly_count:
                    curly_count -= 1
                else:
                    index += 1
                    return index
            elif c == '`':
                done = False
                first = True
                backtick_content = []
                start_index = index
                while not done:
                    m = (RE_TEMPLATE_START if first else RE_TEMPLATE_MIDDLE_END).match(text, index)
                    first = False
                    if m:
                        self.line_num += m.group(0).count('\n')
                        content = self.evaluate_strings(m.group(1), True)
                        if content:
                            backtick_content.append(content)
                        index = m.end(0)
                        if m.group(2) == '${':
                            index = self.find_content(text, index, True)
                        else:
                            done = True
                    else:
                        done = True
                if backtick_content:
                    self.quoted_strings.append([' '.join(backtick_content), self.line_num, 'utf-8'])
            elif c in ('\'', '"'):
                m = RE_STRING.match(text, index)
                if m:
                    self.evaluate_strings(m.group(0)[1:-1])
                    self.line_num += m.group(0).count('\n')
                    index = m.end(0)
            elif c == '\n':
                self.line_num += 1
            elif last == '\n' or c == '/':
                m = RE_COMMENT.match(text, index)
                if m:
                    g = m.groupdict()
                    if g['start'] is None:
                        self.evaluate_inline_tail(g)
                    else:
                        self.evaluate_inline(g)
                    index = m.end(0)
                elif c == '/':
                    m = RE_BLOCK_COMMENT.match(text, index)
                    if m:
                        self.evaluate_block(m.group(0))
                        self.line_num += m.group(0).count('\n')
                        index = m.end(0)

            if index == start_index:
                index += 1
            last = text[index - 1]

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        content = []
        self.jsdoc_comments = []
        self.current_encoding = encoding
        self.line_num = 1
        self.prev_line = -1
        self.leading = ''
        self.block_comments = []
        self.line_comments = []
        self.quoted_strings = []

        self.find_content(self.norm_nl(text))
        self.extend_src(content, context)

        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse JavaScript file."""

        with open(source_file, 'r', encoding=encoding, errors='strict') as f:
            text = f.read()

        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Get filter."""

    return JavaScriptFilter



"""Spell check with Aspell or Hunspell."""
import os
import importlib
from . import util
from .__meta__ import __version__, __version_info__  # noqa: F401
from . import flow_control
from . import filters
from wcmatch import glob
import codecs
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor

__all__ = ("spellcheck",)

STEP_ERROR = """Pipeline step in unexpected format: {}

Each pipeline step should be in the form {{key: options: {{}}}} not {{key: {{}}, key2: {{}}}}
"""


def log(text, level, verbose=0):
    """Log level."""

    if verbose >= level:
        print(text)


class Results(namedtuple('Results', ['words', 'context', 'category', 'error'])):
    """Results."""

    def __new__(cls, words, context, category, error=None):
        """Allow defaults."""

        return super().__new__(cls, words, context, category, error)


class SpellChecker:
    """Spell check class."""

    DICTIONARY = 'dictionary.dic'

    def __init__(self, config, binary='', verbose=0, default_encoding='', debug=False):
        """Initialize."""

        # General options
        self.binary = binary if binary else 'aspell'
        self.verbose = verbose
        self.debug = debug
        self.default_encoding = default_encoding

    def log(self, text, level):
        """Log level."""

        log(text, level, verbose=self.verbose)

    def get_error(self, e):
        """Get the error."""

        import traceback

        return traceback.format_exc() if self.debug else str(e)

    def setup_command(self, encoding, options, personal_dict, file_name=None):
        """Setup the command."""

        return []

    def _pipeline_step(self, sources, options, personal_dict, filter_index=1, flow_status=flow_control.ALLOW):
        """Recursively run text objects through the pipeline steps."""

        for source in sources:
            if source._has_error():
                yield source
            elif not source._is_bytes() and filter_index < len(self.pipeline_steps):
                f = self.pipeline_steps[filter_index]
                if isinstance(f, flow_control.FlowControl):
                    err = ''
                    try:
                        status = f._run(source.category)
                    except Exception as e:
                        err = self.get_error(e)
                        yield filters.SourceText('', source.context, '', '', err)
                    if not err:
                        if filter_index < len(self.pipeline_steps):
                            yield from self._pipeline_step(
                                [source], options, personal_dict, filter_index + 1, status
                            )
                else:
                    if flow_status == flow_control.ALLOW:
                        err = ''
                        try:
                            srcs = f._run(source)
                        except Exception as e:
                            err = self.get_error(e)
                            yield filters.SourceText('', source.context, '', '', err)
                        if not err:
                            yield from self._pipeline_step(
                                srcs, options, personal_dict, filter_index + 1
                            )
                    elif flow_status == flow_control.SKIP:
                        yield from self._pipeline_step(
                            [source], options, personal_dict, filter_index + 1
                        )
                    else:
                        # Halted tasks
                        yield source
            else:
                # Binary content
                yield source

    def _spelling_pipeline(self, sources, options, personal_dict):
        """Check spelling pipeline."""

        for source in self._pipeline_step(sources, options, personal_dict):
            # Don't waste time on empty strings
            if source._has_error():
                yield Results([], source.context, source.category, source.error)
            elif not source.text or source.text.isspace():
                continue
            else:
                encoding = source.encoding
                if source._is_bytes():
                    text = source.text
                else:
                    # UTF-16 and UTF-32 don't work well with Aspell and Hunspell,
                    # so encode with the compatible UTF-8 instead.
                    if encoding.startswith(('utf-16', 'utf-32')):
                        encoding = 'utf-8'
                    text = source.text.encode(encoding)
                self.log('', 3)
                self.log(text, 3)
                cmd = self.setup_command(encoding, options, personal_dict)
                self.log("Command: " + str(cmd), 4)

                try:
                    wordlist = util.call_spellchecker(cmd, input_text=text, encoding=encoding)
                    yield Results(
                        [w for w in sorted(set(wordlist.replace('\r', '').split('\n'))) if w],
                        source.context,
                        source.category
                    )
                except Exception as e:  # pragma: no cover
                    err = self.get_error(e)
                    yield Results([], source.context, source.category, err)

    def spell_check_no_pipeline(self, sources, options, personal_dict):
        """Spell check without the pipeline."""

    def compile_dictionary(self, lang, wordlists, output, verbose):
        """Compile user dictionary."""

    @staticmethod
    def get_options(task):
        """Setup spell checker."""

        return {}

    def setup_dictionary(self, task, binary, verbose):
        """Setup dictionary."""

        return None

    def _build_pipeline(self, task):
        """Build up the pipeline."""

        self.pipeline_steps = []
        kwargs = {}
        if self.default_encoding:
            kwargs["default_encoding"] = self.default_encoding

        steps = task.get('pipeline', [])
        if steps is None:
            self.pipeline_steps = None
        else:
            if not steps:
                steps = task.get('filters', [])
                if steps:
                    util.warn_deprecated(
                        "'filters' key in config is deprecated. 'pipeline' should be used going forward."
                    )

            if not steps:
                steps.append('pyspelling.filters.text')

            for step in steps:
                # Retrieve module and module options
                if isinstance(step, dict):
                    if len(step) > 1:
                        raise ValueError(STEP_ERROR.format(str(step)))
                    name, options = next(iter(step.items()))
                else:
                    name = step
                    options = {}
                if options is None:
                    options = {}

                module = self._get_module(name)
                if issubclass(module, filters.Filter):
                    self.pipeline_steps.append(module(options, **kwargs))
                elif issubclass(module, flow_control.FlowControl):
                    if self.pipeline_steps:
                        self.pipeline_steps.append(module(options))
                    else:
                        raise ValueError("Pipeline cannot start with a 'Flow Control' plugin!")
                else:
                    raise ValueError("'%s' is not a valid plugin!" % name)

    def _get_module(self, module):
        """Get module."""

        if isinstance(module, str):
            mod = importlib.import_module(module)
        for name in ('get_plugin', 'get_filter'):
            attr = getattr(mod, name, None)
            if attr is not None:
                break
            if name == 'get_filter':
                util.warn_deprecated("'get_filter' is deprecated. Plugins should use 'get_plugin'.")
        if not attr:
            raise ValueError("Could not find the 'get_plugin' function in module '%s'!" % module)
        return attr()

    def get_source(self, f):
        """Get the source."""

        if self.pipeline_steps:
            try:
                source = self.pipeline_steps[0]._run_first(f)
            except Exception as e:
                err = self.get_error(e)
                source = [filters.SourceText('', f, '', '', err)]
        else:
            try:
                if self.default_encoding:
                    encoding = filters.PYTHON_ENCODING_NAMES.get(
                        self.default_encoding, self.default_encoding
                    ).lower()
                    encoding = codecs.lookup(encoding).name
                else:
                    encoding = self.default_encoding
                source = [filters.SourceText('', f, encoding, 'file')]
            except Exception as e:
                err = self.get_error(e)
                source = [filters.SourceText('', f, '', '', err)]
        return source


class Aspell(SpellChecker):
    """Aspell spell check class."""

    def __init__(self, config, binary='', verbose=0, default_encoding='', debug=False):
        """Initialize."""

        super().__init__(config, binary, verbose, default_encoding, debug)
        self.binary = binary if binary else 'aspell'

    @staticmethod
    def get_options(task):
        """Setup spell checker."""

        return task.get('aspell', {})

    @classmethod
    def setup_dictionary(cls, task, binary, verbose):
        """Setup dictionary."""

        dictionary_options = task.get('dictionary', {})
        output = os.path.abspath(dictionary_options.get('output', os.path.abspath(cls.DICTIONARY)))
        aspell_options = task.get('aspell', {})
        lang = aspell_options.get('lang', aspell_options.get('l', 'en'))
        wordlists = dictionary_options.get('wordlists', [])
        if lang and wordlists:
            cls.compile_dictionary(
                binary,
                lang,
                dictionary_options.get('wordlists', []),
                dictionary_options.get('encoding', 'utf-8'),
                output,
                verbose
            )
        else:
            output = None
        return output

    @classmethod
    def compile_dictionary(cls, binary, lang, wordlists, encoding, output, verbose):
        """Compile user dictionary."""

        cmd = [
            binary,
            '--lang', lang,
            '--encoding', codecs.lookup(filters.PYTHON_ENCODING_NAMES.get(encoding, encoding).lower()).name,
            'create',
            'master', output
        ]

        wordlist = ''

        try:
            output_location = os.path.dirname(output)
            if not os.path.exists(output_location):
                os.makedirs(output_location)
            if os.path.exists(output):
                os.remove(output)

            log("Compiling Dictionary...", 1, verbose)
            # Read word lists and create a unique set of words
            words = set()
            for wordlist in wordlists:
                with open(wordlist, 'rb') as src:
                    for word in src.read().split(b'\n'):
                        words.add(word.replace(b'\r', b''))

            # Compile wordlist against language
            util.call(
                cmd,
                input_text=b'\n'.join(sorted(words)) + b'\n'
            )
        except Exception:
            log(cmd, 0, verbose)
            log("Current wordlist: '%s'" % wordlist, 0, verbose)
            log("Problem compiling dictionary. Check the binary path and options.", 0, verbose)
            raise

    def spell_check_no_pipeline(self, sources, options, personal_dict):
        """Spell check without the pipeline."""

        for source in sources:

            if source._has_error():  # pragma: no cover
                yield Results([], source.context, source.category, source.error)

            try:
                with open(source.context, 'rb') as f:
                    content = f.read()
            except Exception as e:  # pragma: no cover
                err = self.get_error(e)
                yield Results([], source.context, source.category, err)

            # Don't waste time on empty string
            if not content or content.isspace():
                continue

            self.log('', 3)
            self.log(content, 3)

            cmd = self.setup_command(source.encoding, options, personal_dict, source.context)
            self.log("Command: " + str(cmd), 4)
            try:
                wordlist = util.call_spellchecker(cmd, input_text=content, encoding=source.encoding)
                yield Results(
                    [w for w in sorted(set(wordlist.replace('\r', '').split('\n'))) if w],
                    source.context,
                    source.category
                )
            except Exception as e:  # pragma: no cover
                err = self.get_error(e)
                yield Results([], source.context, source.category, err)

    def setup_command(self, encoding, options, personal_dict, file_name=None):
        """Setup the command."""

        cmd = [
            self.binary,
            'list'
        ]

        if encoding:
            cmd.extend(['--encoding', encoding])

        if personal_dict:
            cmd.extend(['--add-extra-dicts', personal_dict])

        disallowed = {
            '?', 'a', 'c', 'v', 'ignore-repl', 'dont-ignore-repl', 'keyboard', 'prefix', 'repl', 'save-repl',
            'dont-save-repl', 'set-prefix', 'dont-set-prefix', 'sug-mode', 'sug-typo-analysis',
            'dont-sug-typo-analysis', 'sug-repl-table', 'dont-sug-repl-table', 'rem-sug-split-char',
            'add-sug-split-char', 'warn', 'affix-compress', 'dont-affix-compress', 'clean-affixes',
            'dont-clean-affixes', 'invisible-soundslike', 'dont-invisible-soundslike', 'partially-expand',
            'dont-partially-expand', 'skip-invalid-words', 'dont-skip-invalid-words', 'validate-affixes',
            'dont-validate-affixes', 'validate-words', 'dont-validate-words', 'b', 'x', 'backup', 'dont-backup',
            'byte-offsets', 'dont-byte-offsets', 'm', 'P', 'guess', 'dont-guess', 'keymapping', 'reverse',
            'dont-reverse', 'suggest', 'dont-suggest', 'time', 'dont-time'

        }

        if 'mode' not in options:
            options['mode'] = 'none'

        for k, v in options.items():
            if k not in disallowed:
                key = ('-%s' if len(k) == 1 else '--%s') % k
                if isinstance(v, bool) and v is True:
                    cmd.append(key)
                elif isinstance(v, str):
                    cmd.extend([key, v])
                elif isinstance(v, int):
                    cmd.extend([key, str(v)])
                elif isinstance(v, list):
                    for value in v:
                        cmd.extend([key, str(value)])

        if file_name is not None:
            cmd.append(file_name)

        return cmd


class Hunspell(SpellChecker):
    """Hunspell spell check class."""

    def __init__(self, config, binary='', verbose=0, default_encoding='', debug=False):
        """Initialize."""

        super().__init__(config, binary, verbose, default_encoding, debug)
        self.binary = binary if binary else 'hunspell'

    @staticmethod
    def get_options(task):
        """Setup spell checker."""

        return task.get('hunspell', {})

    @classmethod
    def setup_dictionary(cls, task, binary, verbose):
        """Setup dictionary."""

        dictionary_options = task.get('dictionary', {})
        output = os.path.abspath(dictionary_options.get('output', os.path.abspath(cls.DICTIONARY)))
        wordlists = dictionary_options.get('wordlists', [])
        if wordlists:
            cls.compile_dictionary(binary, '', dictionary_options.get('wordlists', []), None, output, verbose)
        else:
            output = None
        return output

    @classmethod
    def compile_dictionary(cls, binary, lang, wordlists, encoding, output, verbose):
        """Compile user dictionary."""

        wordlist = ''

        try:
            output_location = os.path.dirname(output)
            if not os.path.exists(output_location):
                os.makedirs(output_location)
            if os.path.exists(output):
                os.remove(output)

            log("Compiling Dictionary...", 1, verbose)
            # Read word lists and create a unique set of words
            words = set()
            for wordlist in wordlists:
                with open(wordlist, 'rb') as src:
                    for word in src.read().split(b'\n'):
                        words.add(word.replace(b'\r', b''))

            # Sort and create wordlist
            with open(output, 'wb') as dest:
                dest.write(b'\n'.join(sorted(words)) + b'\n')
        except Exception:
            log('Problem compiling dictionary.', 0, verbose)
            log("Current wordlist '%s'" % wordlist, verbose)
            raise

    def spell_check_no_pipeline(self, sources, options, personal_dict):
        """Spell check without the pipeline."""

        for source in sources:
            if source._has_error():  # pragma: no cover
                yield Results([], source.context, source.category, source.error)

            cmd = self.setup_command(source.encoding, options, personal_dict, source.context)
            self.log('', 3)
            self.log("Command: " + str(cmd), 4)
            try:
                wordlist = util.call_spellchecker(cmd, input_text=None, encoding=source.encoding)
                yield Results(
                    [w for w in sorted(set(wordlist.replace('\r', '').split('\n'))) if w],
                    source.context,
                    source.category
                )
            except Exception as e:  # pragma: no cover
                err = self.get_error(e)
                yield Results([], source.context, source.category, err)

    def setup_command(self, encoding, options, personal_dict, file_name=None):
        """Setup command."""

        cmd = [
            self.binary,
            '-l'
        ]

        if encoding:
            cmd.extend(['-i', encoding])

        if personal_dict:
            cmd.extend(['-p', personal_dict])

        disallowed = {
            '1', 'a', 'D', 'G', 'h', 'help', 'l', 'L', 'm', 'P', 's', 'S', 'v', 'vv', 'w'
        }

        for k, v in options.items():
            if k not in disallowed:
                key = ('-%s' if len(k) == 1 else '--%s') % k
                if isinstance(v, bool) and v is True:
                    cmd.append(key)
                elif isinstance(v, str):
                    cmd.extend([key, v])
                elif isinstance(v, int):
                    cmd.extend([key, str(v)])
                elif isinstance(v, list):
                    for value in v:
                        cmd.extend([key, str(value)])

        if file_name is not None:
            cmd.append(file_name)

        return cmd


def iter_tasks(matrix, names, groups):
    """Iterate tasks."""

    # Build name index
    name_index = {task.get('name', ''): index for index, task in enumerate(matrix)}

    for index, task in enumerate(matrix):
        name = task.get('name', '')
        group = task.get('group', '')
        hidden = task.get('hidden', False)
        if names and name in names and index == name_index[name]:
            yield task
        elif groups and group in groups and not hidden:
            yield task
        elif not names and not groups and not hidden:
            yield task


class SpellingTask:
    """Process spelling task."""

    GLOB_FLAG_MAP = {
        "CASE": glob.C,
        "C": glob.C,
        "IGNORECASE": glob.I,
        "I": glob.I,
        "RAWCHARS": glob.R,
        "R": glob.R,
        "NEGATE": glob.N,
        "N": glob.N,
        "MINUSNEGATE": glob.M,
        "M": glob.M,
        "GLOBSTAR": glob.G,
        "G": glob.G,
        "DOTGLOB": glob.D,
        "D": glob.D,
        "EXTGLOB": glob.E,
        "E": glob.E,
        "BRACE": glob.B,
        "B": glob.B,
        "FOLLOW": glob.L,
        "L": glob.L,
        "MATCHBASE": glob.X,
        "X": glob.X,
        "NEGATEALL": glob.A,
        "A": glob.A,
        "NOUNIQUE": glob.Q,
        "Q": glob.Q,
        "GLOBTILDE": glob.T,
        "T": glob.T,
        # We will accept these, but we already force them on
        "SPLIT": glob.S,
        "S": glob.S,
        "NODIR": glob.O,
        "O": glob.O
    }

    def __init__(self, checker, config, binary='', verbose=0, jobs=None, debug=False, skip_dict_compile=False):
        """Initialize."""

        if checker == "hunspell":  # pragma: no cover
            spellchecker = Hunspell
        elif checker == "aspell":
            spellchecker = Aspell
        else:
            raise ValueError('%s is not a valid spellchecker!' % checker)

        self.spellchecker = spellchecker
        self.verbose = verbose
        self.config = config
        self.binary = checker if not binary else binary
        self.debug = debug
        self.jobs = jobs
        self.skip_dict_compile = skip_dict_compile

    def log(self, text, level):
        """Log level."""

        log(text, level, verbose=self.verbose)

    def _to_flags(self, text):
        """Convert text representation of flags to actual flags."""

        flags = 0
        for x in text.split('|'):
            value = x.strip().upper()
            if value:
                flags |= self.GLOB_FLAG_MAP.get(value, 0)
        return flags

    def walk_src(self, targets, flags, limit):
        """Walk source and parse files."""

        for target in targets:
            # Glob using `S` for patterns with `|` and `O` to exclude directories.
            kwargs = {"flags": flags | glob.S | glob.O}
            kwargs['limit'] = limit
            yield from glob.iglob(target, **kwargs)

    def get_checker(self):
        """Get a spell checker object."""

        checker = self.spellchecker(
            self.config,
            self.binary,
            self.verbose,
            self.default_encoding,
            self.debug
        )
        checker._build_pipeline(self.task)
        return checker

    def process_file(self, f, checker):
        """Process a given file."""

        self.log('', 2)
        self.log('> Processing: %s' % f, 1)

        source = checker.get_source(f)

        if checker.pipeline_steps is not None:
            yield from checker._spelling_pipeline(source, self.options, self.personal_dict)
        else:
            yield from checker.spell_check_no_pipeline(source, self.options, self.personal_dict)

    def multi_check(self, f):
        """Check the file for spelling errors (for multi-processing)."""

        return list(self.process_file(f, self.get_checker()))

    def run_task(self, task, source_patterns=None):
        """Walk source and initiate spell check."""

        # Perform spell check
        self.log('Running Task: %s...' % task.get('name', ''), 1)

        # Setup filters and variables for the spell check
        self.task = task
        self.default_encoding = self.task.get('default_encoding', '')
        self.options = self.spellchecker.get_options(self.task)
        if not self.skip_dict_compile:
            self.personal_dict = self.spellchecker.setup_dictionary(self.task, self.binary, self.verbose)
        else:
            dictionary_options = self.task.get('dictionary', {})
            output = os.path.abspath(dictionary_options.get('output', os.path.abspath(self.spellchecker.DICTIONARY)))
            if os.path.exists(output):
                self.personal_dict = output
            else:
                self.personal_dict = self.spellchecker.setup_dictionary(self.task, self.binary, self.verbose)
        self.found_match = False
        glob_flags = self._to_flags(self.task.get('glob_flags', "N|B|G"))
        glob_limit = self.task.get('glob_pattern_limit', 1000)

        if not source_patterns:
            source_patterns = self.task.get('sources', [])

        # If jobs was not specified via command line, check the config for jobs settings
        jobs = self.config.get('jobs', 1) if self.jobs is None else self.jobs

        expect_match = self.task.get('expect_match', True)
        if jobs != 1 and jobs > 0:
            # Use multi-processing to process files concurrently
            with ProcessPoolExecutor(max_workers=jobs if jobs else None) as pool:
                for results in pool.map(self.multi_check, self.walk_src(source_patterns, glob_flags, glob_limit)):
                    self.found_match = True
                    yield from results
        else:
            # Avoid overhead of multiprocessing if we are single threaded
            checker = self.get_checker()
            for f in self.walk_src(source_patterns, glob_flags, glob_limit):
                self.found_match = True
                yield from self.process_file(f, checker)

        if not self.found_match and expect_match:
            raise RuntimeError(
                'None of the source targets from the configuration match any files:\n{}'.format(
                    '\n'.join(f'- {target}' for target in source_patterns)
                )
            )


def spellcheck(
    config_file,
    names=None,
    groups=None,
    binary='',
    checker='',
    sources=None,
    verbose=0,
    debug=False,
    jobs=None,
    skip_dict_compile=False
):
    """Spell check."""

    config = util.read_config(config_file)
    if sources is None:
        sources = []

    matrix = config.get('matrix')
    preferred_checker = config.get('spellchecker', 'aspell')

    if matrix is None:
        matrix = config.get('documents')
        if matrix is not None:
            util.warn_deprecated("'documents' key in config is deprecated. 'matrix' should be used going forward.")
        else:
            raise KeyError(
                'Unable to find or load matrix from pyspelling'
                ' configuration, for more'
                ' details on configuration please read'
                ' https://facelessuser.github.io/pyspelling/configuration/'
            )

    groups = set() if groups is None else set(groups)
    names = set() if names is None else set(names)

    # Sources are only recognized when requesting a single name.
    if (len(names) != 1 and len(sources)):
        sources = []

    processed_tasks = 0

    for task in iter_tasks(matrix, names, groups):

        processed_tasks += 1

        if not checker:
            checker = preferred_checker

        log('Using {} to spellcheck {}'.format(checker, task.get('name', '')), 1, verbose)

        spelltask = SpellingTask(checker, config, binary, verbose, jobs, debug, skip_dict_compile)

        for result in spelltask.run_task(task, source_patterns=sources):
            log('Context: %s' % result.context, 2, verbose)
            yield result

        log("", 1, verbose)

    if processed_tasks == 0:
        raise ValueError(
            'There are either no tasks in the configuration file'
            ' or the specified name or group can not be found.'
        )



"""Wildcard test."""
from .. import util


class TestWildcard(util.PluginTestCase):
    """Test wildcard plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            r"""
            matrix:
            - name: python
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.python:
                  group_comments: true
              - pyspelling.flow_control.wildcard:
                  allow:
                  - py-comment
              - pyspelling.filters.context:
                  context_visible_first: true
                  delimiters:
                  # Ignore lint (noqa) and coverage (pragma) as well as shebang (#!)
                  - open: '^(?: *(?:noqa\b|pragma: no cover)|!)'
                    close: '$'
                  # Ignore Python encoding string -*- encoding stuff -*-
                  - open: '^ *-\*-'
                    close: '-\*-$'
              - pyspelling.filters.context:
                  context_visible_first: true
                  escapes: '\\[\\`]'
                  delimiters:
                  # Ignore multiline content between fences (fences can have 3 or more back ticks)
                  # ```
                  # content
                  # ```
                  - open: '(?s)^(?P<open> *`{{3,}})$'
                    close: '^(?P=open)$'
                  # Ignore text between inline back ticks
                  - open: '(?P<open>`+)'
                    close: '(?P=open)'
            """
        ).format(self.tempdir)
        self.mktemp('.wildcard.yml', config, 'utf-8')

    def test_wildcard(self):
        """Test wildcard."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            #!/usr/bin/env python
            # -*- coding: utf-8 -*-
            \"""
            #! {}
            \"""
            def function():  # noqa
                # ```
                # alsjf alsk
                # eurpoq qeiew
                # ```
            """
        ).format(
            '\n'.join(bad_words + good_words)
        )
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.wildcard.yml', bad_words)



"""Test XML plugin."""
from .. import util


class TestXMLNamespaceNoDefault(util.PluginTestCase):
    """Test XML plugin namespaces without a default namespace."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: xml
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.xml:
                  namespaces:
                    foo: http://me.com/namespaces/foofoo
                    bar: http://me.com/namespaces/foobar
                  ignores:
                  - 'foo|title'
                  - 'bar|*'
                  - '|head'
                  - '*|e2'
                  - 'e3'
            """
        ).format(self.tempdir)
        self.mktemp('.xml.yml', config, 'utf-8')

    def test_xml_namespace_no_default(self):
        """Test XML namespace with no default."""

        template = self.dedent(
            """
            <?xml version="1.0" encoding="UTF-8"?>
            <tag>
              <head>aaaa
              </head>
              <foo:other xmlns:foo="http://me.com/namespaces/foofoo"
                     xmlns:bar="http://me.com/namespaces/foobar">
              <foo:head>
                <foo:title>bbbb</foo:title>
                <bar:title>cccc</bar:title>
              </foo:head>
              <body>
                <foo:e1>dddd</foo:e1>
                <bar:e1>eeee</bar:e1>
                <e1>ffff</e1>
                <foo:e2>gggg</foo:e2>
                <bar:e2>hhhh</bar:e2>
                <e2>iiii</e2>
                <foo:e3>jjjj</foo:e3>
                <bar:e3>kkkk</bar:e3>
                <e3>llll</e3>
              </body>
              </foo:other>
            </tag>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.xml.yml', ['dddd', 'ffff'])


class TestXMLRoot(util.PluginTestCase):
    """Test XML plugin namespaces without a default namespace."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: xml
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.xml:
                  namespaces:
                    foo: http://me.com/namespaces/foofoo
                    bar: http://me.com/namespaces/foobar
                  ignores:
                  - ':root > head'
            """
        ).format(self.tempdir)
        self.mktemp('.xml.yml', config, 'utf-8')

    def test_xml_namespace_no_default(self):
        """Test XML namespace with no default."""

        template = self.dedent(
            """
            <?xml version="1.0" encoding="UTF-8"?>
            <tag>
              <head>aaaa
              </head>
              <foo:other xmlns:foo="http://me.com/namespaces/foofoo"
                     xmlns:bar="http://me.com/namespaces/foobar">
              <foo:head>
                <foo:title>bbbb</foo:title>
                <bar:title>cccc</bar:title>
              </foo:head>
              <body>
                <foo:e1>dddd</foo:e1>
                <bar:e1>eeee</bar:e1>
                <e1>ffff</e1>
                <foo:e2>gggg</foo:e2>
                <bar:e2>hhhh</bar:e2>
                <e2>iiii</e2>
                <foo:e3>jjjj</foo:e3>
                <bar:e3>kkkk</bar:e3>
                <e3>llll</e3>
              </body>
              </foo:other>
            </tag>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck(
            '.xml.yml',
            ['bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll']
        )


class TestXMLNamespace(util.PluginTestCase):
    """Test XML plugin namespaces."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: xml
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.xml:
                  namespaces:
                    "": http://me.com/namespaces/barbar
                    foo: http://me.com/namespaces/foofoo
                    bar: http://me.com/namespaces/foobar
                  ignores:
                  - 'foo|title'
                  - 'bar|*'
                  - '|head'
                  - '*|e2'
                  - e3
            """
        ).format(self.tempdir)
        self.mktemp('.xml.yml', config, 'utf-8')

    def test_xml_namespace(self):
        """Test XML namespace with no default."""

        template = self.dedent(
            """
            <?xml version="1.0" encoding="UTF-8"?>
            <tag xmlns="http://me.com/namespaces/barbar">
              <head>aaaa
              </head>
              <foo:other xmlns:foo="http://me.com/namespaces/foofoo"
                     xmlns:bar="http://me.com/namespaces/foobar">
              <foo:head>
                <foo:title>bbbb</foo:title>
                <bar:title>cccc</bar:title>
              </foo:head>
              <body>
                <foo:e1>dddd</foo:e1>
                <bar:e1>eeee</bar:e1>
                <e1>ffff</e1>
                <foo:e2>gggg</foo:e2>
                <bar:e2>hhhh</bar:e2>
                <e2>iiii</e2>
                <foo:e3>jjjj</foo:e3>
                <bar:e3>kkkk</bar:e3>
                <e3>llll</e3>
              </body>
              </foo:other>
            </tag>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.xml.yml', ['aaaa', 'dddd', 'ffff', 'jjjj'])



"""Version tests."""
import unittest
from pyspelling.__meta__ import Version, parse_version


class TestVersion(unittest.TestCase):
    """Test versions."""

    def test_version_output(self):
        """Test that versions generate proper strings."""

        assert Version(1, 0, 0, "final")._get_canonical() == "1.0"
        assert Version(1, 2, 0, "final")._get_canonical() == "1.2"
        assert Version(1, 2, 3, "final")._get_canonical() == "1.2.3"
        assert Version(1, 2, 0, "alpha", pre=4)._get_canonical() == "1.2a4"
        assert Version(1, 2, 0, "beta", pre=4)._get_canonical() == "1.2b4"
        assert Version(1, 2, 0, "candidate", pre=4)._get_canonical() == "1.2rc4"
        assert Version(1, 2, 0, "final", post=1)._get_canonical() == "1.2.post1"
        assert Version(1, 2, 3, ".dev-alpha", pre=1)._get_canonical() == "1.2.3a1.dev0"
        assert Version(1, 2, 3, ".dev")._get_canonical() == "1.2.3.dev0"
        assert Version(1, 2, 3, ".dev", dev=1)._get_canonical() == "1.2.3.dev1"

    def test_version_comparison(self):
        """Test that versions compare proper."""

        assert Version(1, 0, 0, "final") < Version(1, 2, 0, "final")
        assert Version(1, 2, 0, "alpha", pre=4) < Version(1, 2, 0, "final")
        assert Version(1, 2, 0, "final") < Version(1, 2, 0, "final", post=1)
        assert Version(1, 2, 3, ".dev-beta", pre=2) < Version(1, 2, 3, "beta", pre=2)
        assert Version(1, 2, 3, ".dev") < Version(1, 2, 3, ".dev-beta", pre=2)
        assert Version(1, 2, 3, ".dev") < Version(1, 2, 3, ".dev", dev=1)

    def test_version_parsing(self):
        """Test version parsing."""

        assert parse_version(
            Version(1, 0, 0, "final")._get_canonical()
        ) == Version(1, 0, 0, "final")
        assert parse_version(
            Version(1, 2, 0, "final")._get_canonical()
        ) == Version(1, 2, 0, "final")
        assert parse_version(
            Version(1, 2, 3, "final")._get_canonical()
        ) == Version(1, 2, 3, "final")
        assert parse_version(
            Version(1, 2, 0, "alpha", pre=4)._get_canonical()
        ) == Version(1, 2, 0, "alpha", pre=4)
        assert parse_version(
            Version(1, 2, 0, "beta", pre=4)._get_canonical()
        ) == Version(1, 2, 0, "beta", pre=4)
        assert parse_version(
            Version(1, 2, 0, "candidate", pre=4)._get_canonical()
        ) == Version(1, 2, 0, "candidate", pre=4)
        assert parse_version(
            Version(1, 2, 0, "final", post=1)._get_canonical()
        ) == Version(1, 2, 0, "final", post=1)
        assert parse_version(
            Version(1, 2, 3, ".dev-alpha", pre=1)._get_canonical()
        ) == Version(1, 2, 3, ".dev-alpha", pre=1)
        assert parse_version(
            Version(1, 2, 3, ".dev")._get_canonical()
        ) == Version(1, 2, 3, ".dev")
        assert parse_version(
            Version(1, 2, 3, ".dev", dev=1)._get_canonical()
        ) == Version(1, 2, 3, ".dev", dev=1)

    def test_asserts(self):
        """Test asserts."""

        with self.assertRaises(ValueError):
            Version("1", "2", "3")
        with self.assertRaises(ValueError):
            Version(1, 2, 3, 1)
        with self.assertRaises(ValueError):
            Version("1", "2", "3")
        with self.assertRaises(ValueError):
            Version(1, 2, 3, "bad")
        with self.assertRaises(ValueError):
            Version(1, 2, 3, "alpha")
        with self.assertRaises(ValueError):
            Version(1, 2, 3, "alpha", pre=1, dev=1)
        with self.assertRaises(ValueError):
            Version(1, 2, 3, "alpha", pre=1, post=1)
        with self.assertRaises(ValueError):
            Version(1, 2, 3, ".dev-alpha")
        with self.assertRaises(ValueError):
            Version(1, 2, 3, ".dev-alpha", pre=1, post=1)
        with self.assertRaises(ValueError):
            Version(1, 2, 3, pre=1)
        with self.assertRaises(ValueError):
            Version(1, 2, 3, dev=1)
        with self.assertRaises(ValueError):
            parse_version('bad&version')



"""Text filter."""
from .. import filters
import codecs
import unicodedata


class TextFilter(filters.Filter):
    """Spelling Text."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            'normalize': '',
            'convert_encoding': '',
            'errors': 'strict'
        }

    def validate_options(self, k, v):
        """Validate options."""

        super().validate_options(k, v)
        if k == 'errors' and v.lower() not in ('strict', 'replace', 'ignore', 'backslashreplace'):
            raise ValueError(f"{self.__class__.__name}: '{v}' is not a valid value for '{k}'")
        if k == 'normalize' and v.upper() not in ('NFC', 'NFKC', 'NFD', 'NFKD'):
            raise ValueError(f"{self.__class__.__name}: '{v}' is not a valid value for '{k}'")

    def setup(self):
        """Setup."""

        self.normalize = self.config['normalize'].upper()
        self.convert_encoding = self.config['convert_encoding'].lower()
        self.errors = self.config['errors'].lower()

        if self.convert_encoding:
            self.convert_encoding = codecs.lookup(
                filters.PYTHON_ENCODING_NAMES.get(self.default_encoding, self.default_encoding).lower()
            ).name

            # Don't generate content with BOMs
            if (
                self.convert_encoding.startswith(('utf-32', 'utf-16')) and
                not self.convert_encoding.endswith(('le', 'be'))
            ):
                self.convert_encoding += '-le'

            if self.convert_encoding == 'utf-8-sig':
                self.convert_encoding = 'utf-8'

    def convert(self, text, encoding):
        """Convert the text."""

        if self.normalize in ('NFC', 'NFKC', 'NFD', 'NFKD'):
            text = unicodedata.normalize(self.normalize, text)
        if self.convert_encoding:
            text = text.encode(self.convert_encoding, self.errors).decode(self.convert_encoding)
            encoding = self.convert_encoding
        return text, encoding

    def filter(self, source_file, encoding):  # noqa A001
        """Open and filter the file from disk."""

        with codecs.open(source_file, 'r', encoding=encoding, errors='strict') as f:
            text = f.read()

        text, encoding = self.convert(text, encoding)

        return [filters.SourceText(text, source_file, encoding, 'text')]

    def sfilter(self, source):
        """Execute filter."""

        text, encoding = self.convert(source.text, source.encoding)

        return [filters.SourceText(text, source.context, encoding, 'text')]


def get_plugin():
    """Return the filter."""

    return TextFilter



"""Test HTML plugin."""
from .. import util


class TestHTML(util.PluginTestCase):
    """Test HTML plugin with default `lxml` (HTML) mode."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: html
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  attributes:
                  - alt
                  ignores:
                  - ':is(code, pre)'
                  - 'span:matches(.some-class, #some-id)'
            """
        ).format(self.tempdir)
        self.mktemp('.html.yml', config, 'utf-8')

    def test_html(self):
        """Test HTML."""

        bad_comment_words = ['helo', 'begn']
        bad_content_words = ['flga', 'graet']
        bad_attr_words = ['recieve', 'teh']
        bad_words = bad_comment_words + bad_content_words + bad_attr_words
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            <html>
            <head>
            </head>
            <body>
            <!-- {} -->
            <p>{}<code>kjaljw aljwk</code><img src="./image.png" alt="{}"/></p>
            <pre>uouqei euowuw
            </pre>
            <p><span class="some-class">dksj dkjsk</span><span id="some-id">ksjk akjsks</span>
            </body>
            </html>
            """
        ).format(
            '\n'.join(bad_comment_words + good_words),
            ' '.join(bad_content_words + good_words),
            ' '.join(bad_attr_words + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.html.yml', bad_words)


class TestCSSEscapes(util.PluginTestCase):
    """Test CSS escapes."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
              <h1>A contrived example</h1>
              <svg viewBox="0 0 20 32" class="icon icon-1" data1="1-20-40-50">
                <use xlink:href="images/sprites.svg#icon-undo">aaaa</use>
              </svg>
              <svg viewBox="0 0 30 32" class="icon icon-2" data1="1-20-60-50">
                <use xlink:href:colon="images/sprites.svg#icon-redo">bbbb</use>
              </svg>
              <svg viewBox="0 0 40 32" class="icon icon-3" data="70-20-30-50">
                <use xlink:href="images/sprites.svg#icon-forward">cccc</use>
              </svg>
              <svg viewBox="0 0 50 32" class="icon icon-4" data="100-20-80-50">
                <use xlink:href="other/sprites.svg#icon-reply">dddd</use>
              </svg>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_escapes(self):
        """Test CSS escapes."""

        config = self.dedent(
            """
            matrix:
            - name: html5_css_escapes
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  namespaces:
                    xlink: http://www.w3.org/1999/xlink
                  ignores:
                  - '[xlink\\:href\\:colon^=images]'
                  - '[xlink|\\68r\\65 f^=other]'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck('.html5.yml', ['aaaa', 'cccc'])


class TestHtml5AttrNamespace(util.PluginTestCase):
    """Test HTML plugin HTML5 attribute and namespace logic."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
              <h1>A contrived example</h1>
              <svg viewBox="0 0 20 32" class="icon icon-1" data1="1-20-40-50">
                <use xlink:href="images/sprites.svg#icon-undo">aaaa</use>
              </svg>
              <svg viewBox="0 0 30 32" class="icon icon-2" data1="1-20-60-50">
                <use xlink:href="images/sprites.svg#icon-redo">bbbb</use>
              </svg>
              <svg viewBox="0 0 40 32" class="icon icon-3" data="70-20-30-50">
                <use xlink:href="images/sprites.svg#icon-forward">cccc</use>
              </svg>
              <svg viewBox="0 0 50 32" class="icon icon-4" data="100-20-80-50">
                <use xlink:href="other/sprites.svg#icon-reply">dddd</use>
              </svg>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_html_attr_namespace1(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html5_attr_ns
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  namespaces:
                    xlink: http://www.w3.org/1999/xlink
                  ignores:
                  - '[xlink|href$=undo]'
                  - '[xlink|href^=other]'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck('.html5.yml', ['bbbb', 'cccc'])

    def test_html_attr_namespace2(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html5_attr_ns
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  namespaces:
                    xlink: http://www.w3.org/1999/xlink
                  ignores:
                  - '[xlink|href*=forw]'
                  - '[xlink|href="images/sprites.svg#icon-redo"]'
                  - '[viewbox~="20"]'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck('.html5.yml', ['dddd'])

    def test_html_attr_namespace3(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html5_attr_ns
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  namespaces:
                    xlink: http://www.w3.org/1999/xlink
                  ignores:
                  - '[data1]'
                  - '[data|="100"]'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck('.html5.yml', ['cccc'])


class TestHtml5EmptySelectors(util.PluginTestCase):
    """Test CSS empty selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
            <span>aaaa</span>
            <span> <!-- comment --> </span>
            <span>bbbb</span>
            <span>cccc</span>
            <span>  </span>
            <p>dddd</p>
            <span>eeee</span>
            <span>ffff</span>
            <div>
                <p>gggg</p>
                <span>
                </span>
                <p>hhhh</p>
            </div>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_empty(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'span:empty + *'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'cccc', 'eeee', 'ffff', 'gggg'
            ]
        )


class TestHtml5OnlySelectors(util.PluginTestCase):
    """Test CSS `nth` selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
            <div>
                <p>aaaa</p>
            </div>
            <span>bbbb</span>
            <span>cccc</span>
            <p>dddd</p>
            <span>eeee</span>
            <span>ffff</span>
            <div>
                <p>gggg</p>
                <p>hhhh</p>
            </div>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_only_child(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:only-child'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh'
            ]
        )

    def test_css_only_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:only-of-type'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'eeee', 'ffff', 'gggg', 'hhhh'
            ]
        )


class TestHtml5NthOfSelectors(util.PluginTestCase):
    """Test CSS `nth` selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
            <p>aaaa</p>
            <p>bbbb</p>
            <span class="test">cccc</span>
            <span>dddd</span>
            <span class="test">eeee</span>
            <span>ffff</span>
            <span class="test">gggg</span>
            <p>hhhh</p>
            <p class="test">iiii</p>
            <p>jjjj</p>
            <p class="test">kkkk</p>
            <span>llll</span>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_child_of_s(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - ':nth-child(2n + 1 of :is(p, span).test)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'dddd', 'eeee', 'ffff', 'hhhh', 'iiii', 'jjjj', 'llll'
            ]
        )

    def test_css_child_of_s_vs_schild(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - ':nth-child(-n+3 of p)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(-n+3)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )


class TestHtml5NthSelectors(util.PluginTestCase):
    """Test CSS `nth` selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
            <p>aaaa</p>
            <p>bbbb</p>
            <span>cccc</span>
            <span>dddd</span>
            <span>eeee</span>
            <span>ffff</span>
            <span>gggg</span>
            <p>hhhh</p>
            <p>iiii</p>
            <p>jjjj</p>
            <p>kkkk</p>
            <span>llll</span>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_first_child(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:first-child'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-of-type(2n + 1)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_complex_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - ':nth-of-type(2n + 1):is(p, span)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'dddd', 'ffff', 'iiii', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_odd(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(odd)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'jjjj', 'llll'
            ]
        )

    def test_css_nth_even(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(even)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_child_bad(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(-2)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_child1(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(2)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_child2(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(9n - 1)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_nth_child3(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-child(2n + 1)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'jjjj', 'llll'
            ]
        )

    def test_css_nth_last_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-last-of-type(2n + 1)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'jjjj', 'hhhh', 'llll'
            ]
        )

    def test_css_nth_last_child(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:nth-last-child(2n + 1)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii', 'kkkk', 'llll'
            ]
        )

    def test_css_first_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'span:first-of-type'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_last_child(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'span:last-child'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'kkkk'
            ]
        )

    def test_css_last_type(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:last-of-type'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj', 'llll'
            ]
        )


class TestHtml5AdvancedSelectors(util.PluginTestCase):
    """Test advanced CSS selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
              <div class="aaaa">aaaa
                  <p class="bbbb">bbbb</p>
                  <p class="cccc">cccc</p>
                  <p class="dddd">dddd</p>
                  <div class="eeee">eeee
                  <div class="ffff">ffff
                  <div class="gggg">gggg
                      <p class="hhhh">hhhh</p>
                      <p class="iiii zzzz">iiii</p>
                      <p class="jjjj">jjjj</p>
                      <div class="kkkk">kkkk
                          <p class="llll zzzz">llll</p>
                      </div>
                  </div>
                  </div>
                  </div>
              </div>
            </body>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_css_parents(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'body div.gggg > p.zzzz'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_siblings(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p.bbbb ~ div div.gggg > p.hhhh + p'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'jjjj', 'kkkk', 'llll'
            ]
        )

    def test_css_pseudo(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p.bbbb ~ div > div > div :not(p.hhhh + .zzzz)'
                  - 'p:is(.bbbb + .cccc)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'dddd', 'eeee', 'ffff', 'gggg', 'iiii'
            ]
        )

    def test_css_nested_pseudo(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:is(:not(.zzzz), .iiii)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'eeee', 'ffff', 'gggg', 'kkkk', 'llll'
            ]
        )

    def test_has_pseudo(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'div:not(.aaaa):has(.kkkk > p.llll)'
                  - 'p:has(+ .dddd:has(+ div .jjjj))'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'dddd'
            ]
        )

    def test_has_pseudo2(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'p:has(~ .jjjj)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'jjjj', 'kkkk', 'llll'
            ]
        )


class TestHtml5AdvancedSelectors2(util.PluginTestCase):
    """Test advanced CSS selectors."""

    def setup_fs(self):
        """Setup file system."""

        template = self.dedent(
            """
            <!DOCTYPE html>
            <html>
            <head>
              <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
            </head>
            <body>
              <div class="aaaa">aaaa
                  <p class="bbbb">bbbb</p>
              </div>
              <div class="cccc">cccc
                  <p class="dddd">dddd</p>
              </div>
              <div class="eeee">eeee
                  <p class="ffff">ffff</p>
              </div>
              <div class="gggg">gggg
                  <p class="hhhh">hhhh</p>
              </div>
              <div class="iiii">iiii
                  <p class="jjjj">jjjj</p>
                  <span></span>
              </div>
            </body>
            </body>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')

    def test_has_list(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'div:has(> .bbbb, .ffff, .jjjj)'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'cccc', 'dddd', 'gggg', 'hhhh'
            ]
        )

    def test_not_has_list(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'div:has(> :not(.bbbb, .ffff, .jjjj))'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'eeee', 'ffff'
            ]
        )

    def test_not_has_list2(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'div:not(:has(> .bbbb, .ffff, .jjjj))'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'aaaa', 'bbbb', 'eeee', 'ffff', 'iiii', 'jjjj'
            ]
        )

    def test_not_not(self):
        """Test HTML."""

        config = self.dedent(
            """
            matrix:
            - name: html_css
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  ignores:
                  - 'div:not(:not(.aaaa))'
            """
        ).format(self.tempdir)
        self.mktemp('.html5.yml', config, 'utf-8')
        self.assert_spellcheck(
            '.html5.yml', [
                'cccc', 'dddd', 'eeee', 'ffff', 'gggg', 'hhhh', 'iiii', 'jjjj'
            ]
        )


class TestHTML5LIB(util.PluginTestCase):
    """Test HTML plugin with `html5lib`."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: html
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: html5
                  attributes:
                  - alt
                  ignores:
                  - code
                  - pre
                  - span.some-class
                  - span#some-id
            """
        ).format(self.tempdir)
        self.mktemp('.html5lib.yml', config, 'utf-8')

    def test_html5lib(self):
        """Test `html5lib`."""

        bad_comment_words = ['helo', 'begn']
        bad_content_words = ['flga', 'graet']
        bad_attr_words = ['recieve', 'teh']
        bad_words = bad_comment_words + bad_content_words + bad_attr_words
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            <html>
            <head>
            <script>
            ffffff
            </script>
            </head>
            <body>
            <!-- {} -->
            <p>{}<code>kjaljw aljwk</code><img src="./image.png" alt="{}"/></p>
            <pre>uouqei euowuw
            </pre>
            <p><span class="some-class">dksj dkjsk</span><span id="some-id">ksjk akjsks</span>
            </body>
            </html>
            """
        ).format(
            '\n'.join(bad_comment_words + good_words),
            ' '.join(bad_content_words + good_words),
            ' '.join(bad_attr_words + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.html5lib.yml', bad_words)


class TestXHTML(util.PluginTestCase):
    """Test HTML plugin with `xml` mode."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: html
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  mode: xhtml
                  attributes:
                  - alt
                  ignores:
                  - code
                  - pre
                  - span.some-class
                  - span#some-id
            """
        ).format(self.tempdir)
        self.mktemp('.xhtml.yml', config, 'utf-8')

    def test_xhtml(self):
        """Test XHTML."""

        bad_comment_words = ['helo', 'begn']
        bad_content_words = ['flga', 'graet']
        bad_attr_words = ['recieve', 'teh']
        bad_words = bad_comment_words + bad_content_words + bad_attr_words
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            <?xml version="1.0" encoding="UTF-8"?>
            <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
            <head>
            </head>
            <body>
            <!-- {} -->
            <p>{}<code>kjaljw aljwk</code><img src="./image.png" alt="{}"/></p>
            <pre>uouqei euowuw
            </pre>
            <p><span class="some-class">dksj dkjsk</span><span id="some-id">ksjk akjsks</span></p>
            <pre>
              <div>
                 <p><span><!-- xxxx --></span></p>
              </div>
            </pre>
            </body>
            </html>
            """
        ).format(
            '\n'.join(bad_comment_words + good_words),
            ' '.join(bad_content_words + good_words),
            ' '.join(bad_attr_words + good_words)
        )
        bad_words.append('xxxx')

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.xhtml.yml', bad_words)


class TestHTMLChained(util.PluginTestCase):
    """Test chained HTML plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: html
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.html:
                  attributes:
                  - alt
                  ignores:
                  - code
                  - pre
                  - span.some-class
                  - span#some-id
            """
        ).format(self.tempdir)
        self.mktemp('.html.yml', config, 'utf-8')

    def test_html_after_text(self):
        """Test HTML."""

        bad_comment_words = ['helo', 'begn']
        bad_content_words = ['flga', 'graet']
        bad_attr_words = ['recieve', 'teh']
        bad_words = bad_comment_words + bad_content_words + bad_attr_words
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            <html>
            <head>
            </head>
            <body>
            <!-- {} -->
            <p>{}<code>kjaljw aljwk</code><img src="./image.png" alt="{}"/></p>
            <pre>uouqei euowuw
            </pre>
            <p><span class="some-class">dksj dkjsk</span><span id="some-id">ksjk akjsks</span>
            </body>
            </html>
            """
        ).format(
            '\n'.join(bad_comment_words + good_words),
            ' '.join(bad_content_words + good_words),
            ' '.join(bad_attr_words + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.html.yml', bad_words)


class TestHTMLContext(util.PluginTestCase):
    """Test HTML plugin with default `lxml` (HTML) mode."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: html
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.html:
                  attributes:
                  - alt
                  ignores:
                  - ':is(code, pre)'
                  - 'span:matches(.some-class, #some-id)'
            """
        ).format(self.tempdir)
        self.mktemp('.html.yml', config, 'utf-8')

    def test_html_context(self):
        """Test HTML context."""

        template = self.dedent(
            """
            <html>
            <head>
            </head>
            <body>
            <ol>
            <li>
            <p>Some <code>dkdd</code> text
                flga</p>
                <div><pre><code>kdjk</code></pre></div>
                helo
            </li>
            </ol>
            </html>
            """
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_context(
            '.html.yml',
            [
                'test.txt: html>body>ol>li',
                'test.txt: html>body>ol>li>p'
            ]
        )



"""Test URL plugin."""
from .. import util


class TestURL(util.PluginTestCase):
    """Test URL plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: url
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
                mode: none
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.url:
            """
        ).format(self.tempdir)
        self.mktemp('.url.yml', config, 'utf-8')

    def test_url(self):
        """Test URL."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        text = self.dedent(
            """
            Here is a bad link https://lskjas.com.
            {}
            Here is a bad email slakj.dlaks@mail.com.
            """
        ).format(' '.join(bad_words + good_words))
        self.mktemp('test.txt', text, 'utf-8')
        self.assert_spellcheck('.url.yml', bad_words)


class TestURLChained(util.PluginTestCase):
    """Test chained URL plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: url
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
                d: en_US
                mode: none
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.url:
            """
        ).format(self.tempdir)
        self.mktemp('.url.yml', config, 'utf-8')

    def test_url_after_text(self):
        """Test URL."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        text = self.dedent(
            """
            Here is a bad link https://lskjas.com.
            {}
            Here is a bad email slakj.dlaks@mail.com.
            """
        ).format(' '.join(bad_words + good_words))
        self.mktemp('test.txt', text, 'utf-8')
        self.assert_spellcheck('.url.yml', bad_words)



"""
Office Open XML file format.

Extract text from `docx`, `pptx`, and `xlsx` files.

Specification: http://officeopenxml.com
"""
import io
import bs4
from . import odf
import re
from .. import filters
import soupsieve as sv

DOC_PARAMS = {
    'docx': {
        'filepattern': 'word/{document,header*,footer*,footnotes,endnotes}.xml',
        'namespaces': {'w': 'http://schemas.openxmlformats.org/wordprocessingml/2006/main'},
        'captures': 'w|t'
    },
    'pptx': {
        'filepattern': 'ppt/slides/slide*.xml',
        'namespaces': {'a': 'http://schemas.openxmlformats.org/drawingml/2006/main'},
        'captures': 'a|t'
    },
    'xlsx': {
        'filepattern': 'xl/sharedStrings.xml',
        'namespaces': {'': 'http://schemas.openxmlformats.org/spreadsheetml/2006/main'},
        'captures': 't'
    }
}

RE_SLIDE = re.compile(r'ppt/(?:slides|slideMasters)/(slide(?:Master)?\d+)\.xml', re.I)
RE_DOCS = re.compile(r'word/(document|header\d+|footer\d+|footnotes|endnotes)\.xml')

MIMEMAP = {
    'word': 'docx',
    'xl': 'xlsx',
    'ppt': 'pptx'
}


class OoxmlFilter(odf.OdfFilter):
    """Spelling Python."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {}

    def setup(self):
        """Setup."""

        self.additional_context = ''
        self.comments = False
        self.attributes = []
        self.parser = 'xml'
        self.type = None
        self.filepattern = ''
        self.ignores = None
        self.captures = None

    def has_bom(self, filestream):
        """Check if has BOM."""

        content = filestream.read(4)
        if content == b'PK\x03\x04':
            # Zip file found.
            # Return `BINARY_ENCODE` as content is binary type,
            # but don't return None which means we don't know what we have.
            return filters.BINARY_ENCODE
        # We only handle zip files, so if we are checking this, we've already failed.
        return None

    def content_check(self, filestream):
        """File content check."""

        # We only handle zip files, so if we are checking this, we've already failed.
        return None

    def header_check(self, content):
        """Special encode check."""

        # We only handle zip files, so if we are checking this, we've already failed.
        return None

    def determine_file_type(self, z):
        """Determine file type."""

        content = z.read('[Content_Types].xml')
        with io.BytesIO(content) as b:
            encoding = self._analyze_file(b)
            if encoding is None:
                encoding = 'utf-8'
            b.seek(0)
            text = b.read().decode(encoding)
            soup = bs4.BeautifulSoup(text, 'xml')
            for o in soup.find_all('Override'):
                name = o.attrs.get('PartName')
                for k, v in MIMEMAP.items():
                    if name.startswith(f'/{k}/'):
                        self.type = v
                        break
                if self.type:
                    break
        self.filepattern = DOC_PARAMS[self.type]['filepattern']
        self.namespaces = DOC_PARAMS[self.type]['namespaces']
        self.captures = sv.compile(DOC_PARAMS[self.type]['captures'], DOC_PARAMS[self.type]['namespaces'])

    def soft_break(self, el, text):
        """Apply soft break."""

        # Break word documents by paragraphs.
        if self.type == 'docx' and el.namespace == self.namespaces['w'] and el.name == 'p':
            text.append('\n')
        # Break slides by paragraphs.
        if self.type == 'pptx' and el.namespace == self.namespaces['a'] and el.name == 'p':
            text.append('\n')

    def is_break_tag(self, el):
        """Break on specified boundaries."""

        should_break = False
        return should_break

    def get_context(self, filename):
        """Get context."""

        if self.type == 'pptx':
            context = f'{RE_SLIDE.search(filename).group(1)}: '
        elif self.type == 'docx':
            context = f'{RE_DOCS.match(filename).group(1)}: '
        else:
            context = ''
        return context

    def get_sub_node(self, node):
        """Extract node from document if desired."""

        return node

    def filter(self, source_file, encoding):  # noqa A001
        """Parse XML file."""

        sources = []
        for content, filename, enc in self.get_content(source_file):
            self.additional_context = self.get_context(filename)
            sources.extend(self._filter(content, source_file, enc))
        return sources

    def sfilter(self, source):
        """Filter."""

        sources = []
        for content, filename, enc in self.get_content(io.BytesIO(source.text.encode(source.encoding))):
            self.additional_context = self.get_context(filename)
            sources.extend(self._filter(content, source.context, enc))
        return sources


def get_plugin():
    """Return the filter."""

    return OoxmlFilter



# noqa: A005
"""
XML filter.

Detect encoding from XML header.
"""
from .. import filters
import re
import codecs
import bs4
import soupsieve as sv
from collections import deque, OrderedDict

NON_CONTENT = (bs4.Doctype, bs4.Declaration, bs4.CData, bs4.ProcessingInstruction)

RE_XML_START = re.compile(
    b'^(?:(' +
    b'<\\?xml[^>]+?>' +  # ASCII like
    b')|(' +
    re.escape('<?xml'.encode('utf-32-be')) + b'.+?' + re.escape('>'.encode('utf-32-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-32-le')) + b'.+?' + re.escape('>'.encode('utf-32-le')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-be')) + b'.+?' + re.escape('>'.encode('utf-16-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-le')) + b'.+?' + re.escape('>'.encode('utf-16-le')) +
    b'))'
)

RE_XML_ENCODE = re.compile(br'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')
RE_XML_ENCODE_U = re.compile(r'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')


class XmlFilter(filters.Filter):
    """Spelling Python."""

    default_capture = ['*|*']

    break_tags = set()

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.user_break_tags = set()
        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "comments": True,
            "attributes": [],
            "break_tags": [],
            "ignores": [],
            "captures": self.default_capture,
            "namespaces": {}
        }

    def setup(self):
        """Setup."""

        self.user_break_tags = set(self.config['break_tags'])
        self.comments = self.config['comments']
        self.attributes = set(self.config['attributes'])
        self.parser = 'xml'
        self.type = 'xml'
        ignores = ','.join(self.config['ignores'])
        self.ignores = sv.compile(ignores, self.config['namespaces']) if ignores.strip() else None
        captures = ','.join(self.config['captures'])
        self.captures = sv.compile(captures, self.config['namespaces']) if captures.strip() else None

    def _has_xml_encode(self, content):
        """Check XML encoding."""

        encode = None

        m = RE_XML_START.match(content)
        if m:
            if m.group(1):
                m2 = RE_XML_ENCODE.match(m.group(1))

                if m2:
                    enc = m2.group(2).decode('ascii')

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except LookupError:
                        pass
            else:
                if m.group(2):
                    enc = 'utf-32-be'
                    text = m.group(2)
                elif m.group(3):
                    enc = 'utf-32-le'
                    text = m.group(3)
                elif m.group(4):
                    enc = 'utf-16-be'
                    text = m.group(4)
                elif m.group(5):
                    enc = 'utf-16-le'
                    text = m.group(5)
                try:
                    m2 = RE_XML_ENCODE_U.match(text.decode(enc))
                except Exception:  # pragma: no cover
                    m2 = None

                if m2:
                    enc = m2.group(2)

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except Exception:
                        pass

        return encode

    def header_check(self, content):
        """Special HTML encoding check."""

        return self._has_xml_encode(content)

    def is_break_tag(self, el):
        """Check if tag is an element we should break on."""

        name = el.name
        return name in self.break_tags or name in self.user_break_tags

    def format_blocks(self):
        """Format the text as for a block."""

        block_text = []
        for el, text in self._block_text.items():
            content = ''.join(text)
            if content:
                block_text.append((content, self.construct_selector(el)))
        return block_text

    def construct_selector(self, el, attr=''):
        """Construct an selector for context."""

        selector = deque()

        ancestor = el
        while ancestor and ancestor.parent:
            if ancestor is not el:
                selector.appendleft(ancestor.name)
            else:
                tag = ancestor.name
                prefix = ancestor.prefix
                sel = ''
                if prefix:
                    sel += prefix + '|'
                sel = tag
                if attr:
                    sel += '[%s]' % attr
                selector.appendleft(sel)
            ancestor = ancestor.parent
        return '>'.join(selector)

    def extract_tag_metadata(self, el):
        """Extract meta data."""

    def reset(self):
        """Reset."""

    def get_last_descendant(self, node):
        """Get the last descendant."""

        if node.next_sibling is not None:
            last_descendant = node.next_sibling
        else:
            last_child = node
            while isinstance(last_child, bs4.Tag) and last_child.contents:
                last_child = last_child.contents[-1]
            last_descendant = last_child.next_element

        return last_descendant

    def extract_attributes(self, node):
        """Extract attribute values."""

        for attr in self.attributes:
            value = node.attrs.get(attr, '').strip()
            if value:
                sel = self.construct_selector(node, attr=attr)
                self._attributes.append((value, sel))

    def extract_string(self, node, is_comments):
        """Extract string."""

        string = str(node).strip()
        if string:
            if is_comments:
                sel = self.construct_selector(node.parent) + '<!--comment-->'
                self._comments.append((string, sel))
            else:
                self._block_text[self._current_block].append(string)
                self._block_text[self._current_block].append(' ')

    def pop_block(self, node, force=False):
        """Pop block if the current blocks sibling is found."""

        while self._block_stack and node is self._block_stack[-1][1]:
            self._block_stack.pop(-1)
            self._current_block = self._block_stack[-1][0]

    def set_block(self, node, force=False):
        """Set the current block."""

        self.pop_block(node, force)

        if force or self.is_break_tag(node):
            self._block_stack.append((node, self.get_last_descendant(node)))
            self._block_text[node] = []
            self._current_block = node

    def to_text(self, root):
        """Extract text from the document node."""

        last_capture = None
        last_capture_value = False
        next_good = None

        self._attributes = []
        self._comments = []
        self._block_text = OrderedDict()
        self._block_stack = []
        self.set_block(root, force=True)
        self.extract_tag_metadata(root)

        if not (self.ignores.match(root) if self.ignores else None):
            capture = self.captures.match(root) if self.captures is not None else None
            last_capture = root
            last_capture_value = capture

            if capture:
                self.extract_attributes(root)

            for node in root.descendants:

                if next_good is not None:
                    # We are currently ignoring nodes from an ignored tag.
                    # When we see the first tag that is not a child of the ignored tag,
                    # we can start analyzing tags and text again. Comments are excluded
                    # from ignored, and are captured regardless.
                    if node is not next_good:
                        if self.comments and isinstance(node, bs4.Comment):
                            self.extract_string(node, True)
                        continue
                    next_good = None

                if isinstance(node, bs4.Tag):
                    # Handle tags
                    self.extract_tag_metadata(node)
                    self.set_block(node)

                    if not (self.ignores.match(node) if self.ignores else None):
                        # Handle tags that are not ignored
                        capture = self.captures.match(node) if self.captures is not None else None
                        last_capture = node
                        last_capture_value = capture
                        # Elements that are scheduled to be captured should be checked for attributes to check
                        if capture:
                            self.extract_attributes(node)
                    else:
                        # Handle ignored tags by calculating their last descendant
                        # so we know how long we should ignore nodes
                        next_good = self.get_last_descendant(node)
                        if next_good is None:
                            break
                else:
                    self.pop_block(node)

                    # Handle test nodes: normal text and comments
                    is_comments = isinstance(node, bs4.Comment)
                    if (self.comments and is_comments) or (not is_comments and not isinstance(node, NON_CONTENT)):
                        # Nodes are parsed in order, so sometimes we will descend into another child tag before
                        # all text nodes of the current tag are processed. So we track whether text of the parent
                        # tag should be captured. If we process multiple text nodes of a given tag, we don't
                        # have to recalculate. This allows us to capture text from desired tags.
                        parent = node.parent
                        if is_comments:
                            capture = True
                        elif parent is last_capture:
                            capture = last_capture_value
                        elif not (self.captures.match(parent) if self.captures is not None else None):
                            capture = self.captures.match(parent) if self.captures is not None else None
                            last_capture = parent
                            last_capture_value = capture

                        if capture:
                            self.extract_string(node, is_comments)
        elif self.comments:
            # If the root tag is ignored, but comments is enabled, parse the comments
            for node in root.descendants:
                if isinstance(node, bs4.Comment):
                    self.extract_string(node, True)

        return self.format_blocks(), self._attributes, self._comments

    def _filter(self, text, context, encoding):
        """Filter the source text."""

        content = []
        blocks, attributes, comments = self.to_text(bs4.BeautifulSoup(text, self.parser))
        if self.comments:
            for c, desc in comments:
                content.append(filters.SourceText(c, context + ': ' + desc, encoding, self.type + 'comment'))
        if self.attributes:
            for a, desc in attributes:
                content.append(filters.SourceText(a, context + ': ' + desc, encoding, self.type + 'attribute'))
        for b, desc in blocks:
            content.append(filters.SourceText(b, context + ': ' + desc, encoding, self.type + 'content'))
        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse XML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Return the filter."""

    return XmlFilter
